#define sqrtf (float)sqrt
#include <math.hpp>

#include <d3dx9mesh.h>

//--------------------------------------------------------------------
//  Vertexdefinition
//--------------------------------------------------------------------

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE)
struct Vertex
{
  D3DXVECTOR3 position;
  D3DXVECTOR3 normal;
  D3DCOLOR color;
};

//--------------------------------------------------------------------
//  Klassendeklarationen
//--------------------------------------------------------------------

class Control3D //baut 3D-Device auf und wieder ab
{
  public:
    LPDIRECT3D9 lpDirect3D;
    LPDIRECT3DDEVICE9 lpDevice;
    D3DPRESENT_PARAMETERS presentParameters;
    TForm *lpForm;
  public:
    Control3D(Direct3DParameters p, TForm *_lpForm);
    ~Control3D();
    void adjustProj();
};

class Scene3D //stellt 3D-Szene dar
              // leite hiervon eigene Unterklasse z.B. MyScene3D ab
              // und implementiere dort init, render, move.
              //
              // Fge in der Unterklasse zustzliche Eigenschaften hinzu,
              // die die bentigte 3D-Objekte verwalten knnen, z.B.:
              //
              // Object3D*   spaceship;
              // Terrain3D*  terrain;
              //
              // bzw.
              // Dinge wie Kameraposition, -geschwindigkeit z.B.:
              //
              // D3DXVECTOR3 cameraPos;
              // D3DXVECTOR3 cameraSpeed;
              //
              // etc.
{
  protected:
    Control3D* lpControl3D;
    LPDIRECT3DDEVICE9 lpDevice;

  public:
    Scene3D(Control3D *_lpControl3D);
    void init();   //Daten fr 3D-Darstellung initialisieren
                   //z.B. Kameraposition, Position/Bewegung der Objekte
    void render(); //3D-Szene zeichnen
    void move();   //3D-Szene bewegen
};

class Object3D //verantwortlich fr ein einzelnes 3D-Objekt (aus x-file)
{
  public:
    Control3D* lpControl3D;
    LPDIRECT3DDEVICE9 lpDevice;

    LPD3DXMESH mesh;
    D3DMATERIAL9 *materials;
    DWORD nrMaterials;
    LPDIRECT3DTEXTURE9 *textures;

    D3DXVECTOR3 pos;
    D3DXVECTOR3 speed;
    D3DXVECTOR3 accel;

  public:
    Object3D(Control3D *_lpControl3D);
    Object3D(Control3D *_lpControl3D, char *filename);
    ~Object3D();

    void open(char *filename);

    void draw();
    void move(float timestep);
};

class Triangle3D
{
  public:
    Control3D* lpControl3D;
    LPDIRECT3DDEVICE9 lpDevice;
  private:
    LPDIRECT3DVERTEXBUFFER9 lpVertexBuffer;

  public:
    Triangle3D(Control3D *_lpControl3D);
    ~Triangle3D();
    void draw();
};

class Terrain3D
{
  public:
    Control3D* lpControl3D;
    LPDIRECT3DDEVICE9 lpDevice;

    float** height;
    int nrCol;
    int nrRow;

    float xWidth;
    float zWidth;

    float deltaX;
    float deltaZ;

    float yHeight;
    float variability;

  private:
    LPDIRECT3DVERTEXBUFFER9 lpVertexBuffer;
    LPDIRECT3DINDEXBUFFER9 lpIndexBuffer;
    Vertex* lpVertices;
    WORD* lpIndices;

  public:
    Terrain3D(Control3D *_lpControl3D,
              float _xWidth, float _zWidth,
              int _nrCol, int _nrRow, float _yHeight,
              float _variability);
    ~Terrain3D();
    void draw();
    float getYHeight(float x, float z);
    float getXDerivation(float x, float z);
    float getZDerivation(float x, float z);

  private:
    void calculateHeights(int col1, int row1, int col2, int row2);
};

//--------------------------------------------------------------------
//  Klassenmethoden von Control3D
//--------------------------------------------------------------------

Control3D::Control3D(Direct3DParameters p, TForm *_lpForm)
{
  HRESULT result;

  lpForm = _lpForm;

  lpDirect3D = Direct3DCreate9(D3D_SDK_VERSION);
  if(!lpDirect3D)
  {
    ShowMessage ("Direct3D::Direct3D - Fehler bei Direct3DCreate9");
  }

  ZeroMemory(&presentParameters, sizeof(D3DPRESENT_PARAMETERS));
  presentParameters.BackBufferWidth        = p.displayMode.Width;
  presentParameters.BackBufferHeight       = p.displayMode.Height;
  presentParameters.BackBufferFormat       = p.backBufferFormat;
  presentParameters.BackBufferCount        = 1;
  presentParameters.MultiSampleType        = p.multiSampleType;
  presentParameters.MultiSampleQuality     = p.multiSampleQualityLevel;
  presentParameters.SwapEffect             = D3DSWAPEFFECT_DISCARD;
  presentParameters.hDeviceWindow          = lpForm->Handle;
  presentParameters.Windowed               = p.windowed;
  presentParameters.EnableAutoDepthStencil = true;
  presentParameters.AutoDepthStencilFormat = p.zStencilBufferFormat;
  presentParameters.Flags        = p.zStencilBufferFormat
                                   != D3DFMT_D16_LOCKABLE ?
                                   D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL : 0;
  presentParameters.FullScreen_RefreshRateInHz = p.windowed ?
                                   D3DPRESENT_RATE_DEFAULT :
                                   p.displayMode.RefreshRate;
  presentParameters.PresentationInterval = p.windowed ?
                                   D3DPRESENT_INTERVAL_IMMEDIATE :
                                   D3DPRESENT_INTERVAL_ONE;

  result = lpDirect3D->CreateDevice(p.adapter, p.deviceType,
                                    lpForm->Handle, p.flags,
                                    &presentParameters, &lpDevice);
  if(FAILED(result))
  {
    ShowMessage ("Direct3D::Direct3D - Fehler bei CreateDevice");
  }

  lpDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
  lpDevice->SetRenderState(D3DRS_AMBIENT, 0xffffff);
  lpDevice->SetRenderState(D3DRS_COLORVERTEX, TRUE);
  lpDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
  lpDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);

  adjustProj(); //Projektionsmatrix setzen
}

Control3D::~Control3D()
{
  if(lpDirect3D)
  {
    lpDirect3D->Release();
  }
  if(lpDevice)
  {
    lpDevice->Release();
  }
}

void Control3D::adjustProj()
{
  int width, height;
  float ratio;
  D3DXMATRIX proj;

  if (presentParameters.Windowed)
  {
    width = lpForm->Width;
    height = lpForm->Height;
  }
  else
  {
    width = presentParameters.BackBufferWidth;
    height = presentParameters.BackBufferHeight;
  }

  ratio = ((float)width) / height;

  presentParameters.BackBufferWidth = width;
  presentParameters.BackBufferHeight = height;

  lpDevice->Reset(&presentParameters);

  D3DXMatrixPerspectiveFovLH(&proj, D3DX_PI/4, ratio, 1.0f, 1000.f);
  lpDevice->SetTransform(D3DTS_PROJECTION, &proj);
}

//--------------------------------------------------------------------
//  Klassenmethoden von Scene3D
//--------------------------------------------------------------------

Scene3D::Scene3D(Control3D *_lpControl3D)
{
  lpControl3D = _lpControl3D;
  lpDevice = lpControl3D->lpDevice;
}

//--------------------------------------------------------------------
//  Klassenmethoden von Object3D
//--------------------------------------------------------------------

Object3D::Object3D(Control3D *_lpControl3D)
{
  lpControl3D = _lpControl3D;
  lpDevice = lpControl3D->lpDevice;

  mesh = NULL;
  materials = NULL;
  nrMaterials = 0;
  textures = NULL;

  pos   = D3DXVECTOR3(0.0, 0.0, 0.0);
  speed = D3DXVECTOR3(0.0, 0.0, 0.0);
  accel = D3DXVECTOR3(0.0, 0.0, 0.0);
}

Object3D::Object3D(Control3D *_lpControl3D, char *filename)
{
  lpControl3D = _lpControl3D;
  lpDevice = lpControl3D->lpDevice;

  mesh = NULL;
  materials = NULL;
  nrMaterials = 0;
  textures = NULL;

  open(filename);

  pos   = D3DXVECTOR3(0.0, 0.0, 0.0);
  speed = D3DXVECTOR3(0.0, 0.0, 0.0);
  accel = D3DXVECTOR3(0.0, 0.0, 0.0);
}

void Object3D::open(char *filename)
{
  DWORD i;

  LPD3DXBUFFER buffer;
  D3DXMATERIAL *mat;

  D3DXLoadMeshFromX(filename, D3DXMESH_SYSTEMMEM, lpControl3D->lpDevice,
                    NULL, &buffer, NULL, &nrMaterials, &mesh);

  mat = (D3DXMATERIAL*)buffer->GetBufferPointer();

  materials = new D3DMATERIAL9[nrMaterials];
  textures = new LPDIRECT3DTEXTURE9[nrMaterials];

  for(i=0; i < nrMaterials; i++)
  {
    materials[i] = mat[i].MatD3D;
    materials[i].Ambient.r = 0.2f;
    materials[i].Ambient.g = 0.2f;
    materials[i].Ambient.b = 0.2f;

    materials[i].Diffuse.r = 0.8f;
    materials[i].Diffuse.g = 0.8f;
    materials[i].Diffuse.b = 0.8f;

    materials[i].Emissive.r = 0.0f;
    materials[i].Emissive.g = 0.0f;
    materials[i].Emissive.b = 0.0f;

    materials[i].Specular.r = 1.0f;
    materials[i].Specular.g = 1.0f;
    materials[i].Specular.b = 1.0f;

    materials[i].Power      = 20.0f;

    if(D3DXCreateTextureFromFileEx(lpDevice,
                                   mat[i].pTextureFilename,
                                   D3DX_DEFAULT,
                                   D3DX_DEFAULT,
                                   1, 0,
                                   D3DFMT_UNKNOWN,
                                   D3DPOOL_MANAGED,
                                   D3DX_DEFAULT|D3DX_FILTER_MIRROR_V,
                                   D3DX_FILTER_NONE, 0, NULL, NULL,
                                   &textures[i]) < 0)
    {
      textures[i] = NULL;
    }
  }
  buffer->Release();
}

Object3D::~Object3D()
{
  DWORD i;
  if (materials)
  {
    delete [] materials;
  }
  if (textures)
  {
    for(i=0; i < nrMaterials; i++)
    {
      if (textures[i])
      {
        textures[i]->Release();
      }
      delete [] textures;
    }
  }
  if (mesh)
  {
    mesh->Release();
  }
}

void Object3D::draw()
{
  DWORD i;
  D3DXMATRIX world;

  lpDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
  lpDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
  lpDevice->SetRenderState(D3DRS_SPECULARENABLE, TRUE);
  lpDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);

  D3DXMatrixTranslation(&world, pos.x, pos.y, pos.z);
  lpDevice->SetTransform(D3DTS_WORLD, &world);

  for(i=0; i < nrMaterials; i++)
  {
    lpDevice->SetMaterial(&materials[i]);
    lpDevice->SetTexture(0, textures[i]);
    mesh->DrawSubset(i);
  }
}

void Object3D::move(float timestep)
{
  speed = speed + timestep * accel;
  pos   = pos   + timestep * speed;
}

//--------------------------------------------------------------------
//  Klassenmethoden von Triangle3D
//--------------------------------------------------------------------

Triangle3D::Triangle3D(Control3D *_lpControl3D)
{
  lpControl3D = _lpControl3D;
  lpDevice = lpControl3D->lpDevice;

  lpDevice->CreateVertexBuffer(3*sizeof(Vertex),
                               0,
                               D3DFVF_CUSTOMVERTEX,
                               D3DPOOL_MANAGED,
                               &lpVertexBuffer,
                               NULL);

  Vertex* lpVertices;

  lpVertexBuffer->Lock(0, 0, (void**)&lpVertices, 0);
  lpVertices[0].position = D3DXVECTOR3(-5.0f, 0.0f, 0.0f);
  lpVertices[1].position = D3DXVECTOR3(5.0f, 10.0f, 0.0f);
  lpVertices[2].position = D3DXVECTOR3(5.0f, 0.0f, 10.0f);

  lpVertices[0].normal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
  lpVertices[1].normal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
  lpVertices[2].normal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);

  lpVertices[0].color = D3DCOLOR_XRGB(0,255,0);
  lpVertices[1].color = D3DCOLOR_XRGB(255,0,0);
  lpVertices[2].color = D3DCOLOR_XRGB(0,0,255);
  lpVertexBuffer->Unlock();
}

void Triangle3D::draw()
{
  lpDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
  lpDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
  lpDevice->SetRenderState(D3DRS_SPECULARENABLE, FALSE);

  lpDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);

  lpDevice->SetFVF(D3DFVF_CUSTOMVERTEX);
  lpDevice->SetStreamSource(0, lpVertexBuffer, 0,
                                           sizeof(Vertex));

  lpDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 1);
}

Triangle3D::~Triangle3D()
{
  lpDevice->SetStreamSource(0, NULL, 0, 0);
  lpVertexBuffer->Release();
}

//--------------------------------------------------------------------
//  Klassenmethoden von Terrain3D
//--------------------------------------------------------------------

Terrain3D::Terrain3D(Control3D *_lpControl3D,
                     float _xWidth, float _zWidth,
                     int _nrCol, int _nrRow, float _yHeight,
                     float _variability)
{
  Randomize();

  lpControl3D = _lpControl3D;
  lpDevice = lpControl3D->lpDevice;

  nrCol = _nrCol;
  nrRow = _nrRow;
  xWidth = _xWidth;
  zWidth = _zWidth;
  yHeight = _yHeight;
  variability = _variability;

  deltaX = xWidth / (nrCol - 1);
  deltaZ = zWidth / (nrRow - 1);

  //Einrichten des Hhenfelds
  height = new float* [nrCol];
  for(int col = 0; col < nrCol; col++)
  {
    height[col] = new float[nrRow];
  }

  //zufllige Hhenberechnung
  height[0][0] = RandomRange(-500, 2000) * 0.001f * yHeight;
  height[nrCol-1][0] = RandomRange(-500, 2000) * 0.001f * yHeight;
  height[0][nrRow-1] = RandomRange(-500, 2000) * 0.001f * yHeight;
  height[nrCol-1][nrRow-1] = RandomRange(-500, 2000) * 0.001f * yHeight;
  calculateHeights(0, 0, nrCol - 1, nrRow - 1); //rekursive Hhenberechnung

  //Vertexpuffer mit 3D-Terrainvertices anlegen
  lpDevice->CreateVertexBuffer(nrCol * nrRow *sizeof(Vertex),
                               0,
                               D3DFVF_CUSTOMVERTEX,
                               D3DPOOL_MANAGED,
                               &lpVertexBuffer,
                               NULL);

  lpVertexBuffer->Lock(0, 0, (void**)&lpVertices, 0);

  //Setzen der Vertex-Positionen
  for(int i=0; i < nrCol; i++)
  {
    for(int j=0; j < nrRow; j++)
    {
      if(height[i][j] > 0.0f) //ber Meeresspiegel: grn, Land
      {
        lpVertices[i * nrRow + j].position =
             D3DXVECTOR3(i * deltaX, height[i][j], j * deltaZ);
        lpVertices[i * nrRow + j].color = D3DCOLOR_XRGB(0,255,0);
      }
      else //unter Meeresspiegel: blau, Wasser
      {
        lpVertices[i * nrRow + j].position =
             D3DXVECTOR3(i * deltaX, 0.0f, j * deltaZ);
        lpVertices[i * nrRow + j].color = D3DCOLOR_XRGB(0,0,255);
      }
    }
  }

  //Berechnen der Normalenvektoren (normiert)
  D3DXVECTOR3 vectorA, vectorB1, vectorB2, normal;
  for(int i=0; i < nrCol - 1; i++)
  {
    for(int j=0; j < nrRow - 1; j++)
    {
      vectorA = lpVertices[i * nrRow + j].position;
      vectorB1 = lpVertices[(i+1) * nrRow + j].position - vectorA;
      vectorB2 = lpVertices[i * nrRow + (j+1)].position - vectorA;
      D3DXVec3Cross(&normal, &vectorB2, &vectorB1);
      D3DXVec3Normalize(&lpVertices[i * nrRow + j].normal, &normal);
    }
  }
  for(int i=0; i < nrCol - 1; i++)
  {
    lpVertices[i * nrRow + nrRow - 1].normal = lpVertices[i * nrRow + nrRow - 2].normal;
  }
  for(int j=0; j < nrRow; j++)
  {
    lpVertices[(nrCol - 1) * nrRow + j].normal = lpVertices[(nrCol - 2) * nrRow + j].normal;
  }
  lpVertexBuffer->Unlock();

  //Anlegen der Indizes
  lpDevice->CreateIndexBuffer(2 * (nrCol-1) * nrRow * sizeof(WORD),
                              0,
                              D3DFMT_INDEX16,
                              D3DPOOL_MANAGED,
                              &lpIndexBuffer,
                              NULL);

  lpIndexBuffer->Lock(0, 0, (void**)&lpIndices, 0);
  for(int i=0; i < nrCol - 1; i++)
  {
    for(int j=0; j < nrRow; j++)
    {
      lpIndices[2 * (i * nrRow + j)    ] = i     * nrRow + j;
      lpIndices[2 * (i * nrRow + j) + 1] = (i+1) * nrRow + j;
    }
  }
  lpIndexBuffer->Unlock();
}

void Terrain3D::calculateHeights(int col1, int row1, int col2, int row2)
{
  if ((col1 + 1 < col2) || (row1 + 1 < row2))
  {
    float h11 = height[col1][row1];
    float h21 = height[col2][row1];
    float h12 = height[col1][row2];
    float h22 = height[col2][row2];

    int col = (col1 + col2) / 2;
    int row = (row1 + row2) / 2;

    //Wie gro ist die Teilflche noch im Vergleich zur Gesamtflche?
    //Dieser Anteil bestimmt als factor die noch mgliche Hhenvariation.
    float factor = float(row2 - row1 + 1) * (col2 - col1 + 1) / (nrCol * nrRow);

    if ((col1 < col2) && (row1 + 1 < row2))
    {
      height[col1][row] = 0.5f * (h11 + h12) +
                           RandomRange(-2000, 2000) * 0.001f
                           * variability * factor * yHeight;
      height[col2][row] = 0.5f * (h21 + h22) +
                        RandomRange(-2000, 2000) * 0.001f
                           * variability * factor * yHeight;
    }
    if ((col1 + 1 < col2) && (row1 < row2))
    {
      height[col][row1] = 0.5f * (h11 + h21) +
                        RandomRange(-2000, 2000) * 0.001f
                           * variability * factor * yHeight;
      height[col][row2] = 0.5f * (h12 + h22) +
                        RandomRange(-2000, 2000) * 0.001f
                           * variability * factor * yHeight;
    }
    if ((col1 + 1 < col2) && (row1 + 1 < row2))
    {
      height[col][row] = 0.25f * (h11 + h12 + h21 + h22) +
                        RandomRange(-2000, 2000) * 0.001f
                           * variability * factor * yHeight;
    }
    calculateHeights(col1, row1, col,  row);
    calculateHeights(col,  row1, col2, row);
    calculateHeights(col1, row,  col,  row2);
    calculateHeights(col,  row,  col2, row2);
  }
}

void Terrain3D::draw()
{
  lpDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
  lpDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
  lpDevice->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

  lpDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);

  lpDevice->SetFVF(D3DFVF_CUSTOMVERTEX);
  lpDevice->SetStreamSource(0, lpVertexBuffer, 0,
                                           sizeof(Vertex));
  lpDevice->SetIndices(lpIndexBuffer);

  //lpDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

  //Terrain spaltenweise zeichnen
  for(int i = 0; i < nrRow - 1; i++)
  {
    lpDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP,
                                   0,
                                   2 * nrRow * i,
                                   2 * nrRow,
                                   2 * nrRow * i,
                                   2 * (nrRow + 1));
  }
}

Terrain3D::~Terrain3D()
{
  for(int col = 0; col < nrCol; col++)
  {
    delete [] height[col];
  }                      
  delete [] height;

  lpDevice->SetStreamSource(0, NULL, 0, 0);
  lpDevice->SetIndices(NULL);
  lpVertexBuffer->Release();
  lpIndexBuffer->Release();
}

float Terrain3D::getYHeight(float x, float z)
{
  int i = x / deltaX;
  int j = z / deltaZ;

  float interpolCol = x / deltaX - i;
  float interpolRow = z / deltaZ - j;

  float h11 = height[i][j];
  float h21 = i+1 < nrCol ? height[i+1][j] : h11;
  float h12 = j+1 < nrRow ? height[i][j+1] : h11;
  float h22 = ((i+1 < nrCol) && (j+1 < nrRow)) ? height[i+1][j+1] : h11;

  return (h11 * (1 - interpolCol) * (1 - interpolRow)
          + h21 * interpolCol * (1 - interpolRow)
          + h12 * (1 - interpolCol) * interpolRow
          + h22 * interpolCol * interpolRow);
}

float Terrain3D::getXDerivation(float x, float z)
{
  float dx = 1e-3;
  return ((getYHeight(x+dx, z) - getYHeight(x-dx, z))/(2*dx));
}

float Terrain3D::getZDerivation(float x, float z)
{
  float dz = 1e-3;
  return ((getYHeight(x, z+dz) - getYHeight(x, z-dz))/(2*dz));
}

